/*
 * AUTHOR : K. K. Subramaniam
 * LICENSE : GPL v2
 * Date: 2008-07-12


 * Compile : cc -o sq sq.c
 * Run:
      $ ./sq <squeak.image
 */

#include <stdlib.h>
#include <stdio.h>

#define MAGIC 6502

struct image {
	unsigned magic;
	unsigned headerSize;
	unsigned dataSize;
	unsigned oldBaseAddr;
	unsigned specialObjectsOop;
	unsigned lastHash;
	unsigned savedWindowSize;
	unsigned extraVMMemory;
	unsigned reserved[8];
} header;

/* slots in specials */
#define SP_SCHED	4
#define SP_COMPACT	29

unsigned char *Image;
unsigned specials, compactClasses, sched;

/* slot names of special objects */
char *namesOfSpecials[] = {
	/*  0 */ "header",
		 	/* Undefined and Booleans */
	/*  1 */ "NilObject",
	/*  2 */ "FalseObject",
	/*  3 */ "TrueObject",
		 	/* Root object */
	/*  4 */ "SchedulerAssociation",
			/* Classes */
	/*  5 */ "ClassBitmap",
	/*  6 */ "ClassInteger",
	/*  7 */ "ClassString",
	/*  8 */ "ClassArray",
	/*  9 */ "SystemDictionary",
	/* 10 */ "ClassFloat",
	/* 11 */ "ClassMethodContext",
	/* 12 */ "ClassBlockContext",
	/* 13 */ "ClassPoint",
	/* 14 */ "ClassLargePositiveInteger",
	/* 15 */ "Display",
	/* 16 */ "ClassMessage",
	/* 17 */ "ClassCompiledMethod",
	/* 18 */ "TheLowSpaceSemaphore",
	/* 19 */ "ClassSemaphore",
	/* 20 */ "ClassCharacter",
	/* 21 */ "SelectorDoesNotUnderstand",
	/* 22 */ "SelectorCannotReturn",
	/* 23 */ "ProcessSignalingLowSpace",
	/* 24 */ "SpecialSelectors",
	/* 25 */ "CharacterTable",
	/* 26 */ "SelectorMustBeBoolean",
	/* 27 */ "ClassByteArray",
	/* 28 */ "Process",
	/* 29 */ "CompactClasses",
	/* 30 */ "TheTimerSemaphore",
	/* 31 */ "TheInterruptSemaphore",
	/* 32 */ "Unknown",
	/* 33 */ "Unknown",
	/* 34 */ "Unknown",
	/* 35 */ "SelectorCannotInterpret",
	/* 36 */ "Unknown",
	/* 37 */ "Unknown",
	/* 38 */ "Unknown",
	/* 39 */ "ExternalObjectsArray",
	/* 40 */ "Unknown",
	/* 41 */ "Unknown",
	/* 42 */ "TheFinalizationSemaphore",
	/* 43 */ "ClassLargeNegativeInteger",
	/* 44 */ "ClassExternalAddress",
	/* 45 */ "ClassExternalStructure",
	/* 46 */ "ClassExternalData",
	/* 47 */ "ClassExternalFunction",
	/* 48 */ "ClassExternalLibrary",
	/* 49 */ "SelectorAboutToReturn",
	/* 50 */ "SelectorRunWithIn"
};
#define NSPECIALS 50

unsigned oopAtSlot(unsigned oop, unsigned index)
	/* slot 0 - header */
{
	unsigned *obj = (unsigned *)(Image+oop);

	return obj[index] - header.oldBaseAddr;
}

unsigned valueAtSlot(unsigned oop, unsigned index)
{
	unsigned *obj = (unsigned *)(Image+oop);

	return (obj[index] & 1) ? obj[index] >> 1 : obj[index] - header.oldBaseAddr;
}

unsigned slotsOf(unsigned oop)
{
	unsigned *ptr = (unsigned *)(Image+oop);
	unsigned slots = *ptr >> 2 & 63;

	return slots ? slots : ptr[-2] >> 2;
}

unsigned classOf(unsigned oop)
{
	unsigned *ptr = (unsigned *)(Image+oop);
	unsigned clsi = *ptr >> 12 & 31;

	return clsi ? oopAtSlot(compactClasses, clsi)
			: (ptr[-1] - header.oldBaseAddr) & 0xfffffffc;
}

void printClassAt(unsigned cls)
	/* print class name of class at oop */
{
	unsigned nameoop;

	/* class slots are - header, subclasses, name. name is a ByteSymbol */

	nameoop = valueAtSlot(cls, 2);
}

void printObjectAt(unsigned oop)
{
	unsigned gc, hash, cls, fmt, slots, htype;
	unsigned hdr, *ptr;

	if (oop & 1) {
		printf("oop:%8d val:%d\n", oop, oop >> 1);
		return;
	}
	ptr = (unsigned *)(Image + oop);
	hdr = ptr[0];
	htype = hdr & 3;
	slots = slotsOf(oop);
	fmt   = hdr >> 8 & 15;
	cls = classOf(oop);
	hash  = hdr >> 17 & 4095;
	gc    = hdr >> 29;
	printf("oop:%8d htype:%d cls:%8d hash:%4d fmt:%2d gc:%d slots:%3d\n",
			oop, htype, cls, hash, fmt, gc, slots);
}

main(int argc, char *argv[])
{

	read(0, &header, sizeof(header));
	if (header.magic != MAGIC)
		printf("Warning: magic mismatch: %08x\n", header.magic);
	if (header.headerSize != sizeof(header))
		printf("Warning: header size mismatch: %d %d\n", header.headerSize, sizeof(header));
	specials = header.specialObjectsOop - header.oldBaseAddr;
	printf("magic              : %d\n", header.magic);
	printf("headerSize         : %d\n",   header.headerSize);
	printf("dataSize           : %d\n", header.dataSize);
	printf("oldBaseAddr        : 0x%08x\n", header.oldBaseAddr);
	printf("specialObjectsOop  : oop:%d\n", specials);
	printf("lastHash           : %d\n",   header.lastHash);
	printf("savedWindowSize    : %dx%d\n",  header.savedWindowSize/65536, header.savedWindowSize % 65536);
	printf("extraVMMemory      : %d\n", header.extraVMMemory);

	if ((Image = malloc(header.dataSize)) == (void *)0) {
		perror("sq");
		exit(1);
	}
	if (read(0, Image, header.dataSize) < header.dataSize) {
		perror("sq");
		exit(2);
	}
	/* initialize bootstrap objects */
	compactClasses = oopAtSlot(specials, SP_COMPACT);
	sched = oopAtSlot(specials, SP_SCHED);
	{
		unsigned i, n;
		unsigned proc, context;

		printf("Specials:\n");
		for (i = 1; i <= NSPECIALS; i++) {
			unsigned oop = oopAtSlot(specials, i);
			printf(" %2d%-25s", i-1, namesOfSpecials[i]);
			printObjectAt(oop);
		}
		printf("%-28s", "compact classes");  printObjectAt(compactClasses);
		n = slotsOf(compactClasses);
		for (i=1; i < n; i++) {
			unsigned oop = oopAtSlot(compactClasses, i);
			printf(" %2d ", i);
			printObjectAt(oop);
		}
		printf("%-28s", "scheduler");  printObjectAt(sched);

		/* activeProcess is the second variable in sched */
		proc = oopAtSlot(sched, 2);
		printf("%-28s", "processor");  printObjectAt(proc);

		context = oopAtSlot(proc,1);
		printf("%-28s", "context");  printObjectAt(context);
		printf("%-28s", " sender");  printObjectAt(oopAtSlot(context, 1));
		printf("%-28s", " IP");  printObjectAt(valueAtSlot(context, 2));
		printf("%-28s", " SP");  printObjectAt(valueAtSlot(context, 3));
		printf("%-28s", " Method");  printObjectAt(oopAtSlot(context, 4));
		printf("%-28s", " receiverMap");  printObjectAt(oopAtSlot(context, 5));
		printf("%-28s", " Receiver");  printObjectAt(oopAtSlot(context, 6));
	}
}
